from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .reagent_ref import ReagentRef
from .simple_types import Color, NonNegativeInt, WellID
from .well_sample import WellSample


@dataclass
class Well:
    column: NonNegativeInt
    id: WellID
    row: NonNegativeInt
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    color: Optional[Color] = -1
    external_description: Optional[str] = None
    external_identifier: Optional[str] = None
    reagent_ref: Optional[ReagentRef] = None
    type: Optional[str] = None
    well_sample: List[WellSample] = field(default_factory=list)



from enum import Enum
from typing import Optional

from pydantic.dataclasses import dataclass

from .light_source import LightSource


class Type(Enum):
    HG = "Hg"
    HG_XE = "HgXe"
    OTHER = "Other"
    XE = "Xe"


@dataclass
class Arc(LightSource):
    type: Optional[Type] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .dichroic_ref import DichroicRef
from .filter_ref import FilterRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import FilterSetID

_no_default = object()


@dataclass
class FilterSet(ManufacturerSpec):
    id: FilterSetID = _no_default  # type: ignore
    dichroic_ref: Optional[DichroicRef] = None
    emission_filter_ref: List[FilterRef] = field(default_factory=list)
    excitation_filter_ref: List[FilterRef] = field(default_factory=list)

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")



from dataclasses import field
from enum import Enum
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import FilterID
from .transmittance_range import TransmittanceRange


class Type(Enum):
    BAND_PASS = "BandPass"
    DICHROIC = "Dichroic"
    LONG_PASS = "LongPass"
    MULTI_PASS = "MultiPass"
    NEUTRAL_DENSITY = "NeutralDensity"
    OTHER = "Other"
    SHORT_PASS = "ShortPass"
    TUNEABLE = "Tuneable"


_no_default = object()


@dataclass
class Filter(ManufacturerSpec):
    id: FilterID = _no_default  # type: ignore
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    filter_wheel: Optional[str] = None
    transmittance_range: Optional[TransmittanceRange] = None
    type: Optional[Type] = None

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")



from enum import Enum
from typing import Optional

from pydantic.dataclasses import dataclass

from .simple_types import Hex40


class Compression(Enum):
    BZIP2 = "bzip2"
    NONE = "none"
    ZLIB = "zlib"


@dataclass
class External:
    href: str
    sha1: Hex40
    compression: Optional[Compression] = Compression("none")



from enum import Enum
from typing import Optional

from pydantic.dataclasses import dataclass

from .simple_types import base64Binary


class Compression(Enum):
    BZIP2 = "bzip2"
    NONE = "none"
    ZLIB = "zlib"


_no_default = object()


@dataclass
class BinData(base64Binary):
    big_endian: bool = _no_default  # type: ignore
    length: int = _no_default  # type: ignore
    compression: Optional[Compression] = Compression("none")

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.big_endian is _no_default:
            raise TypeError(
                "__init__ missing 1 required argument: 'big_endian'"
            )
        if self.length is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'length'")



from pydantic.dataclasses import dataclass

from .annotation import Annotation
from .map import Map

_no_default = object()


@dataclass
class MapAnnotation(Annotation):
    value: Map = _no_default  # type: ignore

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.value is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'value'")



from .affine_transform import AffineTransform
from .annotation import Annotation
from .annotation_ref import AnnotationRef
from .arc import Arc
from .basic_annotation import BasicAnnotation
from .bin_data import BinData
from .binary_file import BinaryFile
from .boolean_annotation import BooleanAnnotation
from .channel import Channel
from .channel_ref import ChannelRef
from .comment_annotation import CommentAnnotation
from .dataset import Dataset
from .dataset_ref import DatasetRef
from .detector import Detector
from .detector_settings import DetectorSettings
from .dichroic import Dichroic
from .dichroic_ref import DichroicRef
from .double_annotation import DoubleAnnotation
from .ellipse import Ellipse
from .experiment import Experiment
from .experiment_ref import ExperimentRef
from .experimenter import Experimenter
from .experimenter_group import ExperimenterGroup
from .experimenter_group_ref import ExperimenterGroupRef
from .experimenter_ref import ExperimenterRef
from .external import External
from .filament import Filament
from .file_annotation import FileAnnotation
from .filter import Filter
from .filter_ref import FilterRef
from .filter_set import FilterSet
from .filter_set_ref import FilterSetRef
from .folder import Folder
from .folder_ref import FolderRef
from .generic_excitation_source import GenericExcitationSource
from .image import Image
from .image_ref import ImageRef
from .imaging_environment import ImagingEnvironment
from .instrument import Instrument
from .instrument_ref import InstrumentRef
from .label import Label
from .laser import Laser
from .leader import Leader
from .light_emitting_diode import LightEmittingDiode
from .light_path import LightPath
from .light_source import LightSource
from .light_source_group import LightSourceGroup
from .light_source_settings import LightSourceSettings
from .line import Line
from .list_annotation import ListAnnotation
from .long_annotation import LongAnnotation
from .manufacturer_spec import ManufacturerSpec
from .map import Map
from .map_annotation import MapAnnotation
from .mask import Mask
from .microbeam_manipulation import MicrobeamManipulation
from .microbeam_manipulation_ref import MicrobeamManipulationRef
from .microscope import Microscope
from .numeric_annotation import NumericAnnotation
from .objective import Objective
from .objective_settings import ObjectiveSettings
from .ome import OME
from .pixels import Pixels
from .plane import Plane
from .plate import Plate
from .plate_acquisition import PlateAcquisition
from .point import Point
from .polygon import Polygon
from .polyline import Polyline
from .project import Project
from .project_ref import ProjectRef
from .pump import Pump
from .reagent import Reagent
from .reagent_ref import ReagentRef
from .rectangle import Rectangle
from .reference import Reference
from .rights import Rights
from .roi import ROI
from .roi_ref import ROIRef
from .screen import Screen
from .settings import Settings
from .shape import Shape
from .shape_group import ShapeGroup
from .stage_label import StageLabel
from .structured_annotations import StructuredAnnotations
from .tag_annotation import TagAnnotation
from .term_annotation import TermAnnotation
from .text_annotation import TextAnnotation
from .tiff_data import TiffData
from .timestamp_annotation import TimestampAnnotation
from .transmittance_range import TransmittanceRange
from .type_annotation import TypeAnnotation
from .well import Well
from .well_sample import WellSample
from .well_sample_ref import WellSampleRef


__all__ = [
    "AffineTransform",
    "Annotation",
    "AnnotationRef",
    "Arc",
    "BasicAnnotation",
    "BinData",
    "BinaryFile",
    "BooleanAnnotation",
    "Channel",
    "ChannelRef",
    "CommentAnnotation",
    "Dataset",
    "DatasetRef",
    "Detector",
    "DetectorSettings",
    "Dichroic",
    "DichroicRef",
    "DoubleAnnotation",
    "Ellipse",
    "Experiment",
    "ExperimentRef",
    "Experimenter",
    "ExperimenterGroup",
    "ExperimenterGroupRef",
    "ExperimenterRef",
    "External",
    "Filament",
    "FileAnnotation",
    "Filter",
    "FilterRef",
    "FilterSet",
    "FilterSetRef",
    "Folder",
    "FolderRef",
    "GenericExcitationSource",
    "Image",
    "ImageRef",
    "ImagingEnvironment",
    "Instrument",
    "InstrumentRef",
    "Label",
    "Laser",
    "Leader",
    "LightEmittingDiode",
    "LightPath",
    "LightSource",
    "LightSourceGroup",
    "LightSourceSettings",
    "Line",
    "ListAnnotation",
    "LongAnnotation",
    "ManufacturerSpec",
    "Map",
    "MapAnnotation",
    "Mask",
    "MicrobeamManipulation",
    "MicrobeamManipulationRef",
    "Microscope",
    "NumericAnnotation",
    "OME",
    "Objective",
    "ObjectiveSettings",
    "Pixels",
    "Plane",
    "Plate",
    "PlateAcquisition",
    "Point",
    "Polygon",
    "Polyline",
    "Project",
    "ProjectRef",
    "Pump",
    "ROI",
    "ROIRef",
    "Reagent",
    "ReagentRef",
    "Rectangle",
    "Reference",
    "Rights",
    "Screen",
    "Settings",
    "Shape",
    "ShapeGroup",
    "StageLabel",
    "StructuredAnnotations",
    "TagAnnotation",
    "TermAnnotation",
    "TextAnnotation",
    "TiffData",
    "TimestampAnnotation",
    "TransmittanceRange",
    "TypeAnnotation",
    "Well",
    "WellSample",
    "WellSampleRef",
]



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .experimenter_ref import ExperimenterRef
from .light_source_settings import LightSourceSettings
from .roi_ref import ROIRef
from .simple_types import MicrobeamManipulationID


@dataclass
class MicrobeamManipulation:
    experimenter_ref: ExperimenterRef
    id: MicrobeamManipulationID
    roi_ref: List[ROIRef]
    description: Optional[str] = None
    light_source_settings: List[LightSourceSettings] = field(
        default_factory=list
    )
    type = None



from typing import Optional

from pydantic.dataclasses import dataclass


@dataclass
class ManufacturerSpec:
    lot_number: Optional[str] = None
    manufacturer: Optional[str] = None
    model: Optional[str] = None
    serial_number: Optional[str] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .experimenter_ref import ExperimenterRef
from .microbeam_manipulation import MicrobeamManipulation
from .simple_types import ExperimentID


@dataclass
class Experiment:
    id: ExperimentID
    description: Optional[str] = None
    experimenter_ref: Optional[ExperimenterRef] = None
    microbeam_manipulation: List[MicrobeamManipulation] = field(
        default_factory=list
    )
    type = None



from enum import Enum
from typing import Optional

from pydantic.dataclasses import dataclass

from .light_source import LightSource
from .pump import Pump
from .simple_types import (
    PositiveFloat,
    PositiveInt,
    UnitsFrequency,
    UnitsLength,
)


class LaserMedium(Enum):
    AG = "Ag"
    ALEXANDRITE = "Alexandrite"
    AR = "Ar"
    AR_CL = "ArCl"
    AR_FL = "ArFl"
    CO = "CO"
    CO2 = "CO2"
    COUMARIN_C30 = "CoumarinC30"
    CU = "Cu"
    E_MINUS = "EMinus"
    ER_GLASS = "ErGlass"
    ER_YAG = "ErYAG"
    GA_AL_AS = "GaAlAs"
    GA_AS = "GaAs"
    H2_O = "H2O"
    H_FL = "HFl"
    HE_CD = "HeCd"
    HE_NE = "HeNe"
    HO_YAG = "HoYAG"
    HO_YLF = "HoYLF"
    KR = "Kr"
    KR_CL = "KrCl"
    KR_FL = "KrFl"
    N = "N"
    ND_GLASS = "NdGlass"
    ND_YAG = "NdYAG"
    OTHER = "Other"
    RHODAMINE6_G = "Rhodamine6G"
    RUBY = "Ruby"
    TI_SAPPHIRE = "TiSapphire"
    XE = "Xe"
    XE_BR = "XeBr"
    XE_CL = "XeCl"
    XE_FL = "XeFl"


class Pulse(Enum):
    CW = "CW"
    MODE_LOCKED = "ModeLocked"
    OTHER = "Other"
    Q_SWITCHED = "QSwitched"
    REPETITIVE = "Repetitive"
    SINGLE = "Single"


class Type(Enum):
    DYE = "Dye"
    EXCIMER = "Excimer"
    FREE_ELECTRON = "FreeElectron"
    GAS = "Gas"
    METAL_VAPOR = "MetalVapor"
    OTHER = "Other"
    SEMICONDUCTOR = "Semiconductor"
    SOLID_STATE = "SolidState"


@dataclass
class Laser(LightSource):
    frequency_multiplication: Optional[PositiveInt] = None
    laser_medium: Optional[LaserMedium] = None
    pockel_cell: Optional[bool] = None
    pulse: Optional[Pulse] = None
    pump: Optional[Pump] = None
    repetition_rate: Optional[float] = None
    repetition_rate_unit: Optional[UnitsFrequency] = UnitsFrequency("Hz")
    tuneable: Optional[bool] = None
    type: Optional[Type] = None
    wavelength: Optional[PositiveFloat] = None
    wavelength_unit: Optional[UnitsLength] = UnitsLength("nm")



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .simple_types import ReagentID


@dataclass
class Reagent:
    id: ReagentID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None
    name: Optional[str] = None
    reagent_identifier: Optional[str] = None



from typing import Optional

from pydantic.dataclasses import dataclass

from .simple_types import (
    NonNegativeFloat,
    PercentFraction,
    PositiveFloat,
    UnitsLength,
)


@dataclass
class TransmittanceRange:
    cut_in: Optional[PositiveFloat] = None
    cut_in_tolerance: Optional[NonNegativeFloat] = None
    cut_in_tolerance_unit: Optional[UnitsLength] = UnitsLength("nm")
    cut_in_unit: Optional[UnitsLength] = UnitsLength("nm")
    cut_out: Optional[PositiveFloat] = None
    cut_out_tolerance: Optional[NonNegativeFloat] = None
    cut_out_tolerance_unit: Optional[UnitsLength] = UnitsLength("nm")
    cut_out_unit: Optional[UnitsLength] = UnitsLength("nm")
    transmittance: Optional[PercentFraction] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .folder_ref import FolderRef
from .image_ref import ImageRef
from .roi_ref import ROIRef
from .simple_types import FolderID


@dataclass
class Folder:
    id: FolderID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None
    folder_ref: List[FolderRef] = field(default_factory=list)
    image_ref: List[ImageRef] = field(default_factory=list)
    name: Optional[str] = None
    roi_ref: List[ROIRef] = field(default_factory=list)



from dataclasses import field
from enum import Enum
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .detector_settings import DetectorSettings
from .filter_set_ref import FilterSetRef
from .light_path import LightPath
from .light_source_settings import LightSourceSettings
from .simple_types import (
    ChannelID,
    Color,
    PositiveFloat,
    PositiveInt,
    UnitsLength,
)


class IlluminationType(Enum):
    EPIFLUORESCENCE = "Epifluorescence"
    NON_LINEAR = "NonLinear"
    OBLIQUE = "Oblique"
    OTHER = "Other"
    TRANSMITTED = "Transmitted"


class ContrastMethod(Enum):
    BRIGHTFIELD = "Brightfield"
    DARKFIELD = "Darkfield"
    DIC = "DIC"
    FLUORESCENCE = "Fluorescence"
    HOFFMAN_MODULATION = "HoffmanModulation"
    OBLIQUE_ILLUMINATION = "ObliqueIllumination"
    OTHER = "Other"
    PHASE = "Phase"
    POLARIZED_LIGHT = "PolarizedLight"


class AcquisitionMode(Enum):
    BRIGHT_FIELD = "BrightField"
    FLUORESCENCE_CORRELATION_SPECTROSCOPY = (
        "FluorescenceCorrelationSpectroscopy"
    )
    FLUORESCENCE_LIFETIME = "FluorescenceLifetime"
    FSM = "FSM"
    LASER_SCANNING_CONFOCAL_MICROSCOPY = "LaserScanningConfocalMicroscopy"
    LCM = "LCM"
    MULTI_PHOTON_MICROSCOPY = "MultiPhotonMicroscopy"
    NEAR_FIELD_SCANNING_OPTICAL_MICROSCOPY = (
        "NearFieldScanningOpticalMicroscopy"
    )
    OTHER = "Other"
    PALM = "PALM"
    SECOND_HARMONIC_GENERATION_IMAGING = "SecondHarmonicGenerationImaging"
    SINGLE_MOLECULE_IMAGING = "SingleMoleculeImaging"
    SLIT_SCAN_CONFOCAL = "SlitScanConfocal"
    SPECTRAL_IMAGING = "SpectralImaging"
    SPIM = "SPIM"
    SPINNING_DISK_CONFOCAL = "SpinningDiskConfocal"
    STED = "STED"
    STORM = "STORM"
    STRUCTURED_ILLUMINATION = "StructuredIllumination"
    SWEPT_FIELD_CONFOCAL = "SweptFieldConfocal"
    TIRF = "TIRF"
    TOTAL_INTERNAL_REFLECTION = "TotalInternalReflection"
    WIDE_FIELD = "WideField"


@dataclass
class Channel:
    id: ChannelID
    acquisition_mode: Optional[AcquisitionMode] = None
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    color: Optional[Color] = -1
    contrast_method: Optional[ContrastMethod] = None
    detector_settings: Optional[DetectorSettings] = None
    emission_wavelength: Optional[PositiveFloat] = None
    emission_wavelength_unit: Optional[UnitsLength] = UnitsLength("nm")
    excitation_wavelength: Optional[PositiveFloat] = None
    excitation_wavelength_unit: Optional[UnitsLength] = UnitsLength("nm")
    filter_set_ref: Optional[FilterSetRef] = None
    fluor: Optional[str] = None
    illumination_type: Optional[IlluminationType] = None
    light_path: Optional[LightPath] = None
    light_source_settings: Optional[LightSourceSettings] = None
    name: Optional[str] = None
    nd_filter: Optional[float] = None
    pinhole_size: Optional[float] = None
    pinhole_size_unit: Optional[UnitsLength] = UnitsLength("µm")
    pockel_cell_setting: Optional[int] = None
    samples_per_pixel: Optional[PositiveInt] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .reagent import Reagent
from .reference import Reference
from .simple_types import PlateID, ScreenID

_no_default = object()


@dataclass
class PlateRef(Reference):
    id: PlateID = _no_default  # type: ignore

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")


@dataclass
class Screen:
    id: ScreenID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None
    name: Optional[str] = None
    plate_ref: List[PlateRef] = field(default_factory=list)
    protocol_description: Optional[str] = None
    protocol_identifier: Optional[str] = None
    reagent: List[Reagent] = field(default_factory=list)
    reagent_set_description: Optional[str] = None
    reagent_set_identifier: Optional[str] = None
    type: Optional[str] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .detector import Detector
from .dichroic import Dichroic
from .filter import Filter
from .filter_set import FilterSet
from .light_source import LightSource
from .light_source_group import LightSourceGroup
from .microscope import Microscope
from .objective import Objective
from .simple_types import InstrumentID


@dataclass
class Instrument:
    id: InstrumentID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    detector: List[Detector] = field(default_factory=list)
    dichroic: List[Dichroic] = field(default_factory=list)
    filter: List[Filter] = field(default_factory=list)
    filter_set: List[FilterSet] = field(default_factory=list)
    light_source_group: List[LightSourceGroup] = field(default_factory=list)
    microscope: Optional[Microscope] = None
    objective: List[Objective] = field(default_factory=list)



from typing import Optional

from pydantic.dataclasses import dataclass

from .settings import Settings
from .simple_types import (
    Binning,
    DetectorID,
    PositiveInt,
    UnitsElectricPotential,
    UnitsFrequency,
)

_no_default = object()


@dataclass
class DetectorSettings(Settings):
    id: DetectorID = _no_default  # type: ignore
    binning: Optional[Binning] = None
    gain: Optional[float] = None
    integration: Optional[PositiveInt] = None
    offset: Optional[float] = None
    read_out_rate: Optional[float] = None
    read_out_rate_unit: Optional[UnitsFrequency] = UnitsFrequency("MHz")
    voltage: Optional[float] = None
    voltage_unit: Optional[UnitsElectricPotential] = UnitsElectricPotential(
        "V"
    )
    zoom: Optional[float] = None

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")



from dataclasses import field
from enum import Enum
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import ObjectiveID, UnitsLength


class Immersion(Enum):
    AIR = "Air"
    GLYCEROL = "Glycerol"
    MULTI = "Multi"
    OIL = "Oil"
    OTHER = "Other"
    WATER = "Water"
    WATER_DIPPING = "WaterDipping"


class Correction(Enum):
    ACHRO = "Achro"
    ACHROMAT = "Achromat"
    APO = "Apo"
    FL = "Fl"
    FLUAR = "Fluar"
    FLUOR = "Fluor"
    FLUOTAR = "Fluotar"
    NEOFLUAR = "Neofluar"
    OTHER = "Other"
    PLAN_APO = "PlanApo"
    PLAN_FLUOR = "PlanFluor"
    PLAN_NEOFLUAR = "PlanNeofluar"
    SUPER_FLUOR = "SuperFluor"
    UV = "UV"
    VIOLET_CORRECTED = "VioletCorrected"


_no_default = object()


@dataclass
class Objective(ManufacturerSpec):
    id: ObjectiveID = _no_default  # type: ignore
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    calibrated_magnification: Optional[float] = None
    correction: Optional[Correction] = None
    immersion: Optional[Immersion] = None
    iris: Optional[bool] = None
    lens_na: Optional[float] = None
    nominal_magnification: Optional[float] = None
    working_distance: Optional[float] = None
    working_distance_unit: Optional[UnitsLength] = UnitsLength("µm")

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

from ome_autogen import convert_schema
from .schema import to_dict, validate

__all__ = ["convert_schema", "to_dict", "validate", "from_xml"]


try:
    from .model import OME
except ImportError:
    print("OME dataclasses not found ... running autogeneration")
    import runpy

    runpy.run_module("ome_autogen", run_name="__main__")
    from .model import OME


def from_xml(xml, OME=OME) -> OME:

    d = to_dict(xml)
    for key in list(d.keys()):
        if key.startswith(("xml", "xsi")):
            d.pop(key)

    return OME(**d)



import builtins
import os
import re
import shutil
from typing import Generator, Iterable, List, Set, Union

import black
import isort
from xmlschema import XMLSchema, qnames
from xmlschema.validators import (
    XsdAnyAttribute,
    XsdAnyElement,
    XsdAttribute,
    XsdElement,
    XsdType,
)

# FIXME: hacks
OVERRIDE = {
    "MetadataOnly": ("bool", "False", None),
    "XMLAnnotation": ("Optional[str]", "None", "from typing import Optional\n\n",),
    "BinData/Length": ("int", None, None),
    "ROI/Union": (
        "conlist(Shape, min_items=1)",
        None,
        "from pydantic.types import conlist\nfrom .shape import Shape",
    ),
    "TiffData/UUID": (
        r'Optional[dataclass(type("UUID", (), {"__annotations__": '
        r'{"file_name": str, "value": UniversallyUniqueIdentifier}}))]',
        None,
        "from typing import Optional\n\nfrom "
        ".simple_types import UniversallyUniqueIdentifier",
    ),
}


def black_format(text, line_length=79):
    return black.format_str(text, mode=black.FileMode(line_length=line_length))


def sort_imports(text):
    return isort.SortImports(file_contents=text).output


def sort_types(el):
    if not el.is_complex() and not el.base_type.is_restriction():
        return "    " + el.local_name.lower()
    return el.local_name.lower()


def sort_prop(prop):
    return ("" if prop.default_val_str else "   ") + prop.format().lower()


def as_identifier(s):
    # Remove invalid characters
    _s = re.sub("[^0-9a-zA-Z_]", "", s)
    # Remove leading characters until we find a letter or underscore
    _s = re.sub("^[^a-zA-Z_]+", "", _s)
    if not _s:
        raise ValueError(f"Could not clean {s}: nothing left")
    return _s


def camel_to_snake(name):
    # https://stackoverflow.com/a/1176023
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower().replace(" ", "_")


def local_import(item_type):
    return f"from .{camel_to_snake(item_type)} import {item_type}"


def make_dataclass(component) -> List[str]:
    lines = ["from pydantic.dataclasses import dataclass", ""]
    if isinstance(component, XsdType):
        base_type = component.base_type
    else:
        base_type = component.type.base_type

    if base_type and not hasattr(base_type, "python_type"):
        base_name = f"({base_type.local_name})"
        if base_type.is_complex():
            lines += [local_import(base_type.local_name)]
        else:
            lines += [f"from .simple_types import {base_type.local_name}"]
    else:
        base_name = ""

    base_members = set()
    _basebase = base_type
    while _basebase:
        base_members.update(set(iter_members(base_type)))
        _basebase = _basebase.base_type

    members = MemberSet(m for m in iter_members(component) if m not in base_members)
    lines += members.imports()
    lines += members.locals()

    cannot_have_required_args = base_type and members.has_non_default_args()
    if cannot_have_required_args:
        lines += ["_no_default = object()", ""]

    lines += ["@dataclass", f"class {component.local_name}{base_name}:"]
    lines += members.lines(
        indent=1,
        force_defaults=" = _no_default  # type: ignore"
        if cannot_have_required_args
        else None,
    )

    if cannot_have_required_args:
        lines += ["", "    # hack for dataclass inheritance with non-default args"]
        lines += ["    # https://stackoverflow.com/a/53085935/"]
        lines += ["    def __post_init__(self):"]
        for m in members.non_defaults:
            lines += [f"        if self.{m.identifier} is _no_default:"]
            lines += [
                "    " * 3 + 'raise TypeError("__init__ missing 1 '
                f'required argument: {m.identifier!r}")'
            ]

    return lines


def make_enum(component, name=None):
    name = name or component.local_name
    lines = ["from enum import Enum", ""]
    lines += [f"class {name}(Enum):"]
    enum_elems = list(component.elem.iter("enum"))
    facets = component.get_facet(qnames.XSD_ENUMERATION)
    members = []
    if enum_elems:
        for el, value in zip(enum_elems, facets.enumeration):
            name = el.attrib["enum"]
            if component.base_type.python_type.__name__ == "str":
                value = f'"{value}"'
            members.append((name, value))
    else:
        for e in facets.enumeration:
            members.append((camel_to_snake(e), repr(e)))

    for n, v in sorted(members):
        lines.append(f"    {as_identifier(n).upper()} = {v}")
    return lines


facet_converters = {
    qnames.XSD_PATTERN: lambda f: [f"regex = re.compile(r'{f.regexps[0]}')"],
    qnames.XSD_MIN_INCLUSIVE: lambda f: [f"ge = {f.value}"],
    qnames.XSD_MIN_EXCLUSIVE: lambda f: [f"gt = {f.value}"],
    qnames.XSD_MAX_INCLUSIVE: lambda f: [f"le = {f.value}"],
    qnames.XSD_MAX_EXCLUSIVE: lambda f: [f"lt = {f.value}"],
    qnames.XSD_LENGTH: lambda f: [f"min_length = {f.value}", f"max_length = {f.value}"],
    qnames.XSD_MIN_LENGTH: lambda f: [f"min_length = {f.value}"],
    qnames.XSD_MAX_LENGTH: lambda f: [f"max_length = {f.value}"],
}


def iter_all_members(component):
    for c in component.iter_components((XsdElement, XsdAttribute)):
        if c is component:
            continue
        yield c


def iter_members(
    component: Union[XsdElement, XsdType]
) -> Generator[Union[XsdElement, XsdAttribute], None, None]:
    if isinstance(component, XsdElement):
        for attr in component.attributes.values():
            if isinstance(attr, XsdAttribute):
                yield attr
        for elem in component.iterchildren():
            yield elem
    else:
        yield from iter_all_members(component)


class Member:
    def __init__(self, component: Union[XsdElement, XsdAttribute]):
        self.component = component
        assert not component.is_global()

    @property
    def identifier(self) -> str:
        if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
            return self.component.local_name
        ident = camel_to_snake(self.component.local_name)
        if not ident.isidentifier():
            raise ValueError(f"failed to make identifier of {self!r}")
        return ident

    @property
    def type(self) -> XsdType:
        return self.component.type

    @property
    def is_enum_type(self) -> bool:
        return self.type.get_facet(qnames.XSD_ENUMERATION) is not None

    @property
    def is_builtin_type(self) -> bool:
        return hasattr(self.type, "python_type")

    @property
    def key(self):
        p = self.component.parent
        name = p.local_name
        while not name and (p.parent is not None):
            p = p.parent
            name = p.local_name
        name = f"{name}/{self.component.local_name}"
        if name not in OVERRIDE and self.component.local_name in OVERRIDE:
            return self.component.local_name
        return name

    def locals(self) -> Set[str]:
        if self.key in OVERRIDE:
            return set()
        if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
            return set()
        if not self.type or self.type.is_global():
            return set()
        locals_: Set[str] = set()
        # FIXME: this bit is mostly hacks
        if self.type.is_complex() and self.component.ref is None:
            locals_.add("\n".join(make_dataclass(self.component)) + "\n")
        if self.type.is_restriction() and self.is_enum_type:
            locals_.add(
                "\n".join(make_enum(self.type, name=self.component.local_name)) + "\n"
            )
        return locals_

    def imports(self) -> Set[str]:
        if self.key in OVERRIDE:
            _imp = OVERRIDE[self.key][2]
            return set([_imp]) if _imp else set()
        if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
            return set(["from typing import Any"])
        imports = set()
        if not self.max_occurs:
            imports.add("from typing import List")
            if self.is_optional:
                imports.add("from dataclasses import field")
        elif self.is_optional:
            imports.add("from typing import Optional")
        if self.type.is_datetime():
            imports.add("from datetime import datetime")
        if not self.is_builtin_type and self.type.is_global():
            # FIXME: hack
            if not self.type.local_name == "anyType":
                if self.type.is_complex():
                    imports.add(local_import(self.type.local_name))
                else:
                    imports.add(f"from .simple_types import {self.type.local_name}")

        if self.component.ref is not None:
            if self.component.ref.local_name not in OVERRIDE:
                imports.add(local_import(self.component.ref.local_name))

        return imports

    @property
    def type_string(self) -> str:
        """single type, without Optional, etc..."""
        if self.key in OVERRIDE:
            return OVERRIDE[self.key][0]
        if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
            return "Any"
        if self.component.ref is not None:
            assert self.component.ref.is_global()
            return self.component.ref.local_name

        if self.type.is_datetime():
            return "datetime"
        if self.is_builtin_type:
            return self.type.python_type.__name__

        if self.type.is_global():
            return self.type.local_name
        elif self.type.is_complex():
            return self.component.local_name

        if self.type.is_restriction():
            # enumeration
            enum = self.type.get_facet(qnames.XSD_ENUMERATION)
            if enum:
                return self.component.local_name
            if self.type.base_type.local_name == "string":
                return "str"
        return ""

    @property
    def full_type_string(self) -> str:
        """full type, like Optional[List[str]]"""
        if self.key in OVERRIDE and self.type_string:
            return f": {self.type_string}"
        type_string = self.type_string
        if not type_string:
            return ""
        if not self.max_occurs:
            type_string = f"List[{type_string}]"
        elif self.is_optional:
            type_string = f"Optional[{type_string}]"
        return f": {type_string}" if type_string else ""

    @property
    def default_val_str(self) -> str:
        if self.key in OVERRIDE:
            default = OVERRIDE[self.key][1]
            return f" = {default}" if default else ""
        if not self.is_optional:
            return ""

        if not self.max_occurs:
            default_val = "field(default_factory=list)"
        else:
            default_val = self.component.default
            if default_val is not None:
                if self.is_enum_type:
                    default_val = f"{self.type_string}('{default_val}')"
                elif hasattr(builtins, self.type_string):
                    default_val = repr(getattr(builtins, self.type_string)(default_val))
            else:
                default_val = "None"
        return f" = {default_val}"

    @property
    def max_occurs(self) -> bool:
        return getattr(self.component, "max_occurs", 1)

    @property
    def is_optional(self) -> bool:
        # FIXME: hack.  doesn't fully capture the restriction
        if getattr(self.component.parent, "model", "") == "choice":
            return True
        if hasattr(self.component, "min_occurs"):
            return self.component.min_occurs == 0
        return self.component.is_optional()

    def __repr__(self) -> str:
        type_ = "element" if isinstance(self.component, XsdElement) else "attribute"
        return f"<Member {type_} {self.component.local_name}>"

    def format(self, force_default=None) -> str:
        default = self.default_val_str
        if force_default:
            default = default or force_default
        return f"{self.identifier}{self.full_type_string}{default}"


class MemberSet:
    def __init__(self, initial: Iterable = ()):
        self._members: Set[Member] = set()
        self.update(initial)

    def add(self, member: Member):
        if not isinstance(member, Member):
            member = Member(member)
        self._members.add(member)

    def update(self, members: Iterable):
        for member in members:
            self.add(member)

    def lines(self, indent=1, force_defaults: str = None) -> List[str]:
        if not self._members:
            lines = ["    " * indent + "pass"]
        else:
            lines = [
                "    " * indent + m.format(force_defaults)
                for m in sorted(self._members, key=lambda x: sort_prop(x))
            ]
        return lines

    def imports(self) -> List[str]:
        if self._members:
            return list(set.union(*[m.imports() for m in self._members]))
        return []

    def locals(self) -> List[str]:
        if self._members:
            return list(set.union(*[m.locals() for m in self._members]))
        return []

    def has_non_default_args(self) -> bool:
        return any(not m.default_val_str for m in self._members)

    @property
    def non_defaults(self) -> "MemberSet":
        return MemberSet(m for m in self._members if not m.default_val_str)

    def __iter__(self):
        return iter(self._members)


class GlobalElem:
    def __init__(self, elem: Union[XsdElement, XsdType]):
        assert elem.is_global()
        self.elem = elem

    @property
    def type(self):
        return self.elem if self.is_type else self.elem.type

    @property
    def is_complex(self):
        if hasattr(self.type, "is_complex"):
            return self.type.is_complex()
        return False

    @property
    def is_element(self):
        return isinstance(self.elem, XsdElement)

    @property
    def is_type(self):
        return isinstance(self.elem, XsdType)

    @property
    def is_enum(self) -> bool:
        is_enum = bool(self.elem.get_facet(qnames.XSD_ENUMERATION) is not None)
        if is_enum:
            if not len(self.elem.facets) == 1:
                raise NotImplementedError("Unexpected enum with multiple facets")
        return is_enum

    def _simple_class(self) -> List[str]:
        if self.is_enum:
            return make_enum(self.elem)

        lines = []
        if self.type.base_type.is_restriction():
            parent = self.type.base_type.local_name
        else:
            # it's a restriction of a builtin
            pytype = self.elem.base_type.python_type.__name__
            parent = f"Constrained{pytype.title()}"
            lines.extend([f"from pydantic.types import {parent}", ""])
        lines.append(f"class {self.elem.local_name}({parent}):")

        members = []
        for key, facet in self.elem.facets.items():
            members.extend([f"    {line}" for line in facet_converters[key](facet)])
        lines.extend(members if members else ["    pass"])
        if any("re.compile" in m for m in members):
            lines = ["import re", ""] + lines
        return lines

    def _abstract_class(self) -> List[str]:
        # FIXME: ? this might be a bit of an OME-schema-specific hack
        # this seems to be how abstract is used in the OME schema
        for e in self.elem.iter_components():
            if e != self.elem:
                raise NotImplementedError(
                    "Don't yet know how to handle abstract class with sub-components"
                )

        subs = [
            el
            for el in self.elem.schema.elements.values()
            if el.substitution_group == self.elem.name
        ]

        if not subs:
            raise NotImplementedError(
                "Don't know how to handle abstract class without substitutionGroups"
            )

        for el in subs:
            if not el.type.is_extension() and el.type.base_type == self.elem.type:
                raise NotImplementedError(
                    "Expected all items in substitution group to extend "
                    f"the type {self.elem.type} of Abstract element {self.elem}"
                )

        sub_names = [el.local_name for el in subs]
        lines = ["from typing import Union"]
        lines.extend([local_import(n) for n in sub_names])
        lines += [local_import(self.elem.type.local_name)]
        lines += [f"{self.elem.local_name} = {self.elem.type.local_name}", ""]
        lines += [f"{self.elem.local_name}Type = Union[{', '.join(sub_names)}]"]
        return lines

    def lines(self) -> str:
        if not self.is_complex:
            lines = self._simple_class()
        elif self.elem.abstract:
            lines = self._abstract_class()
        else:
            lines = make_dataclass(self.elem)
        return "\n".join(lines)

    def format(self) -> str:
        return black_format(sort_imports(self.lines() + "\n"))

    def write(self, filename: str) -> None:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as f:
            f.write(self.format())

    @property
    def fname(self) -> str:
        return f"{camel_to_snake(self.elem.local_name)}.py"


_this_dir = os.path.dirname(__file__)
# _url = os.path.join(_this_dir, "ome_types", "ome-2016-06.xsd")
_url = "https://www.openmicroscopy.org/Schemas/OME/2016-06/ome.xsd"
_target = os.path.join(_this_dir, "ome_types", "model")


def convert_schema(url=_url, target_dir=_target):
    print("Inspecting XML schema ...")
    schema = XMLSchema(url)
    print("Building dataclasses ...")
    shutil.rmtree(target_dir, ignore_errors=True)
    init_imports = []
    simples: List[GlobalElem] = []
    for elem in sorted(schema.types.values(), key=sort_types):
        if elem.local_name in OVERRIDE:
            continue
        converter = GlobalElem(elem)
        if not elem.is_complex():
            simples.append(converter)
            continue
        targetfile = os.path.join(target_dir, converter.fname)
        init_imports.append((converter.fname, elem.local_name))
        converter.write(filename=targetfile)

    for elem in schema.elements.values():
        if elem.local_name in OVERRIDE:
            continue
        converter = GlobalElem(elem)
        targetfile = os.path.join(target_dir, converter.fname)
        init_imports.append((converter.fname, elem.local_name))
        converter.write(filename=targetfile)

    text = "\n".join([s.format() for s in simples])
    text = black_format(sort_imports(text))
    with open(os.path.join(target_dir, "simple_types.py"), "w") as f:
        f.write(text)

    text = ""
    for fname, classname in init_imports:
        text += local_import(classname) + "\n"
    text = sort_imports(text)
    text += f"\n\n__all__ = [{', '.join(sorted(repr(i[1]) for i in init_imports))}]"
    text = black_format(text)
    with open(os.path.join(target_dir, f"__init__.py"), "w") as f:
        f.write(text)


if __name__ == "__main__":
    # for testing
    convert_schema()



import pickle
import re
from os.path import dirname, exists, join
from typing import Dict

import xmlschema

__cache__: Dict[str, xmlschema.XMLSchema] = {}


def camel_to_snake(name):
    # https://stackoverflow.com/a/1176023
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower().replace(" ", "_")


def get_schema(xml):
    url = xmlschema.fetch_schema(xml)
    version = (
        re.split("(.com|.org)", url)[-1]
        .replace("/", "_")
        .lstrip("_")
        .replace(".xsd", "")
    )
    if version not in __cache__:
        local = join(dirname(__file__), f"{version}.pkl")
        if exists(local):
            with open(local, "rb") as f:
                __cache__[version] = pickle.load(f)
        else:
            __cache__[version] = xmlschema.XMLSchema(url)
            with open(local, "wb") as f:
                pickle.dump(__cache__[version], f)
    return __cache__[version]


def validate(xml: str, schema=None):
    schema = schema or get_schema(xml)
    schema.validate(xml)


class MyConverter(xmlschema.converters.XMLSchemaConverter):
    def __init__(self, namespaces=None):
        super().__init__(namespaces, attr_prefix="")

    def map_qname(self, qname):
        name = super().map_qname(qname)
        return camel_to_snake(name)

    def element_decode(self, data, xsd_element, xsd_type=None, level=0) -> dict:
        """Converts a decoded element data to a data structure."""
        result = super().element_decode(data, xsd_element, xsd_type, level)
        if result and "$" in result:
            result["value"] = result.pop("$")
        return result


def to_dict(xml: str, schema=None, converter=MyConverter, **kwargs):
    schema = schema or get_schema(xml)
    return schema.to_dict(xml, converter=converter, **kwargs)



from typing import Optional

from pydantic.dataclasses import dataclass

from .shape import Shape
from .simple_types import Marker

_no_default = object()


@dataclass
class Line(Shape):
    x1: float = _no_default  # type: ignore
    x2: float = _no_default  # type: ignore
    y1: float = _no_default  # type: ignore
    y2: float = _no_default  # type: ignore
    marker_end: Optional[Marker] = None
    marker_start: Optional[Marker] = None

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.y2 is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'y2'")
        if self.x1 is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'x1'")
        if self.x2 is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'x2'")
        if self.y1 is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'y1'")



from dataclasses import field
from enum import Enum
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .bin_data import BinData
from .channel import Channel
from .plane import Plane
from .simple_types import (
    PixelsID,
    PixelType,
    PositiveFloat,
    PositiveInt,
    UnitsLength,
    UnitsTime,
)
from .tiff_data import TiffData


class DimensionOrder(Enum):
    XYCTZ = "XYCTZ"
    XYCZT = "XYCZT"
    XYTCZ = "XYTCZ"
    XYTZC = "XYTZC"
    XYZCT = "XYZCT"
    XYZTC = "XYZTC"


@dataclass
class Pixels:
    dimension_order: DimensionOrder
    id: PixelsID
    size_c: PositiveInt
    size_t: PositiveInt
    size_x: PositiveInt
    size_y: PositiveInt
    size_z: PositiveInt
    type: PixelType
    big_endian: Optional[bool] = None
    bin_data: List[BinData] = field(default_factory=list)
    channel: List[Channel] = field(default_factory=list)
    interleaved: Optional[bool] = None
    metadata_only: bool = False
    physical_size_x: Optional[PositiveFloat] = None
    physical_size_x_unit: Optional[UnitsLength] = UnitsLength("µm")
    physical_size_y: Optional[PositiveFloat] = None
    physical_size_y_unit: Optional[UnitsLength] = UnitsLength("µm")
    physical_size_z: Optional[PositiveFloat] = None
    physical_size_z_unit: Optional[UnitsLength] = UnitsLength("µm")
    plane: List[Plane] = field(default_factory=list)
    significant_bits: Optional[PositiveInt] = None
    tiff_data: List[TiffData] = field(default_factory=list)
    time_increment: Optional[float] = None
    time_increment_unit: Optional[UnitsTime] = UnitsTime("s")



from dataclasses import field
from datetime import datetime
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .experiment_ref import ExperimentRef
from .experimenter_group_ref import ExperimenterGroupRef
from .experimenter_ref import ExperimenterRef
from .imaging_environment import ImagingEnvironment
from .instrument_ref import InstrumentRef
from .microbeam_manipulation_ref import MicrobeamManipulationRef
from .objective_settings import ObjectiveSettings
from .pixels import Pixels
from .roi_ref import ROIRef
from .simple_types import ImageID
from .stage_label import StageLabel


@dataclass
class Image:
    id: ImageID
    pixels: Pixels
    acquisition_date: Optional[datetime] = None
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None
    experiment_ref: Optional[ExperimentRef] = None
    experimenter_group_ref: Optional[ExperimenterGroupRef] = None
    experimenter_ref: Optional[ExperimenterRef] = None
    imaging_environment: Optional[ImagingEnvironment] = None
    instrument_ref: Optional[InstrumentRef] = None
    microbeam_manipulation_ref: List[MicrobeamManipulationRef] = field(
        default_factory=list
    )
    name: Optional[str] = None
    objective_settings: Optional[ObjectiveSettings] = None
    roi_ref: List[ROIRef] = field(default_factory=list)
    stage_label: Optional[StageLabel] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .text_annotation import TextAnnotation

_no_default = object()


@dataclass
class TagAnnotation(TextAnnotation):
    value: str = _no_default  # type: ignore
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.value is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'value'")



from pydantic.dataclasses import dataclass

from .reference import Reference
from .simple_types import WellSampleID

_no_default = object()


@dataclass
class WellSampleRef(Reference):
    id: WellSampleID = _no_default  # type: ignore

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")



from dataclasses import field
from enum import Enum
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import DetectorID, UnitsElectricPotential


class Type(Enum):
    ANALOG_VIDEO = "AnalogVideo"
    APD = "APD"
    CCD = "CCD"
    CMOS = "CMOS"
    CORRELATION_SPECTROSCOPY = "CorrelationSpectroscopy"
    EBCCD = "EBCCD"
    EMCCD = "EMCCD"
    FTIR = "FTIR"
    INTENSIFIED_CCD = "IntensifiedCCD"
    LIFETIME_IMAGING = "LifetimeImaging"
    OTHER = "Other"
    PHOTODIODE = "Photodiode"
    PMT = "PMT"
    SPECTROSCOPY = "Spectroscopy"


_no_default = object()


@dataclass
class Detector(ManufacturerSpec):
    id: DetectorID = _no_default  # type: ignore
    amplification_gain: Optional[float] = None
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    gain: Optional[float] = None
    offset: Optional[float] = None
    type: Optional[Type] = None
    voltage: Optional[float] = None
    voltage_unit: Optional[UnitsElectricPotential] = UnitsElectricPotential(
        "V"
    )
    zoom: Optional[float] = None

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")



from typing import Optional

from pydantic.dataclasses import dataclass

from .simple_types import UnitsLength


@dataclass
class StageLabel:
    name: str
    x: Optional[float] = None
    x_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    y: Optional[float] = None
    y_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    z: Optional[float] = None
    z_unit: Optional[UnitsLength] = UnitsLength("reference frame")



from datetime import datetime
from typing import Optional

from pydantic.dataclasses import dataclass

from .image_ref import ImageRef
from .simple_types import NonNegativeInt, UnitsLength, WellSampleID


@dataclass
class WellSample:
    id: WellSampleID
    index: NonNegativeInt
    image_ref: Optional[ImageRef] = None
    position_x: Optional[float] = None
    position_x_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    position_y: Optional[float] = None
    position_y_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    timepoint: Optional[datetime] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .simple_types import ExperimenterID


@dataclass
class Experimenter:
    id: ExperimenterID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    email: Optional[str] = None
    first_name: Optional[str] = None
    institution: Optional[str] = None
    last_name: Optional[str] = None
    middle_name: Optional[str] = None
    user_name: Optional[str] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .experimenter_ref import ExperimenterRef
from .leader import Leader
from .simple_types import ExperimenterGroupID


@dataclass
class ExperimenterGroup:
    id: ExperimenterGroupID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None
    experimenter_ref: List[ExperimenterRef] = field(default_factory=list)
    leader: List[Leader] = field(default_factory=list)
    name: Optional[str] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass


@dataclass
class M:
    k: Optional[str] = None


@dataclass
class Map:
    k: Optional[str] = None
    m: List[M] = field(default_factory=list)



from typing import Optional

from pydantic.dataclasses import dataclass

from .light_source import LightSource
from .map import Map


@dataclass
class GenericExcitationSource(LightSource):
    map: Optional[Map] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .plate_acquisition import PlateAcquisition
from .simple_types import (
    NamingConvention,
    NonNegativeInt,
    PlateID,
    PositiveInt,
    UnitsLength,
)
from .well import Well


@dataclass
class Plate:
    id: PlateID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    column_naming_convention: Optional[NamingConvention] = None
    columns: Optional[PositiveInt] = None
    description: Optional[str] = None
    external_identifier: Optional[str] = None
    field_index: Optional[NonNegativeInt] = None
    name: Optional[str] = None
    plate_acquisition: List[PlateAcquisition] = field(default_factory=list)
    row_naming_convention: Optional[NamingConvention] = None
    rows: Optional[PositiveInt] = None
    status: Optional[str] = None
    well: List[Well] = field(default_factory=list)
    well_origin_x: Optional[float] = None
    well_origin_x_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    well_origin_y: Optional[float] = None
    well_origin_y_unit: Optional[UnitsLength] = UnitsLength("reference frame")



from typing import Optional

from pydantic.dataclasses import dataclass

from .settings import Settings
from .simple_types import (
    LightSourceID,
    PercentFraction,
    PositiveFloat,
    UnitsLength,
)

_no_default = object()


@dataclass
class LightSourceSettings(Settings):
    id: LightSourceID = _no_default  # type: ignore
    attenuation: Optional[PercentFraction] = None
    wavelength: Optional[PositiveFloat] = None
    wavelength_unit: Optional[UnitsLength] = UnitsLength("nm")

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")



from pydantic.dataclasses import dataclass

from .shape import Shape

_no_default = object()


@dataclass
class Rectangle(Shape):
    height: float = _no_default  # type: ignore
    width: float = _no_default  # type: ignore
    x: float = _no_default  # type: ignore
    y: float = _no_default  # type: ignore

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.height is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'height'")
        if self.width is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'width'")
        if self.y is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'y'")
        if self.x is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'x'")



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .dataset_ref import DatasetRef
from .experimenter_group_ref import ExperimenterGroupRef
from .experimenter_ref import ExperimenterRef
from .simple_types import ProjectID


@dataclass
class Project:
    id: ProjectID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    dataset_ref: List[DatasetRef] = field(default_factory=list)
    description: Optional[str] = None
    experimenter_group_ref: Optional[ExperimenterGroupRef] = None
    experimenter_ref: Optional[ExperimenterRef] = None
    name: Optional[str] = None



from typing import Optional

from pydantic.dataclasses import dataclass

from .bin_data import BinData
from .external import External
from .simple_types import NonNegativeLong


@dataclass
class BinaryFile:
    file_name: str
    size: NonNegativeLong
    bin_data: Optional[BinData] = None
    external: Optional[External] = None
    mime_type: Optional[str] = None



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .dichroic_ref import DichroicRef
from .filter_ref import FilterRef


@dataclass
class LightPath:
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    dichroic_ref: Optional[DichroicRef] = None
    emission_filter_ref: List[FilterRef] = field(default_factory=list)
    excitation_filter_ref: List[FilterRef] = field(default_factory=list)



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .simple_types import AnnotationID, ExperimenterID


@dataclass
class Annotation:
    id: AnnotationID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    annotator: Optional[ExperimenterID] = None
    description: Optional[str] = None
    namespace: Optional[str] = None



from typing import Optional

from pydantic.dataclasses import dataclass

from .map import Map
from .simple_types import PercentFraction, UnitsPressure, UnitsTemperature


@dataclass
class ImagingEnvironment:
    air_pressure: Optional[float] = None
    air_pressure_unit: Optional[UnitsPressure] = UnitsPressure("mbar")
    co2_percent: Optional[PercentFraction] = None
    humidity: Optional[PercentFraction] = None
    map: Optional[Map] = None
    temperature: Optional[float] = None
    temperature_unit: Optional[UnitsTemperature] = UnitsTemperature("°C")



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass
from pydantic.types import conlist

from .annotation_ref import AnnotationRef
from .shape import Shape
from .simple_types import ROIID


@dataclass
class ROI:
    id: ROIID
    union: conlist(Shape, min_items=1)
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None
    name: Optional[str] = None



from enum import Enum
from typing import Optional

from pydantic.dataclasses import dataclass

from .manufacturer_spec import ManufacturerSpec


class Type(Enum):
    DISSECTION = "Dissection"
    ELECTROPHYSIOLOGY = "Electrophysiology"
    INVERTED = "Inverted"
    OTHER = "Other"
    UPRIGHT = "Upright"


@dataclass
class Microscope(ManufacturerSpec):
    type: Optional[Type] = None



from typing import Optional

from pydantic.dataclasses import dataclass

from .shape import Shape
from .simple_types import Marker

_no_default = object()


@dataclass
class Polyline(Shape):
    points: str = _no_default  # type: ignore
    marker_end: Optional[Marker] = None
    marker_start: Optional[Marker] = None

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.points is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'points'")



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .dataset import Dataset
from .experiment import Experiment
from .experimenter import Experimenter
from .experimenter_group import ExperimenterGroup
from .folder import Folder
from .image import Image
from .instrument import Instrument
from .plate import Plate
from .project import Project
from .rights import Rights
from .roi import ROI
from .screen import Screen
from .simple_types import UniversallyUniqueIdentifier
from .structured_annotations import StructuredAnnotations


@dataclass
class BinaryOnly:
    metadata_file: str
    uuid: UniversallyUniqueIdentifier


@dataclass
class OME:
    binary_only: Optional[BinaryOnly] = None
    creator: Optional[str] = None
    dataset: List[Dataset] = field(default_factory=list)
    experiment: List[Experiment] = field(default_factory=list)
    experimenter: List[Experimenter] = field(default_factory=list)
    experimenter_group: List[ExperimenterGroup] = field(default_factory=list)
    folder: List[Folder] = field(default_factory=list)
    image: List[Image] = field(default_factory=list)
    instrument: List[Instrument] = field(default_factory=list)
    plate: List[Plate] = field(default_factory=list)
    project: List[Project] = field(default_factory=list)
    rights: Optional[Rights] = None
    roi: List[ROI] = field(default_factory=list)
    screen: List[Screen] = field(default_factory=list)
    structured_annotations: Optional[StructuredAnnotations] = None
    uuid: Optional[UniversallyUniqueIdentifier] = None



from enum import Enum
from typing import Optional

from pydantic.dataclasses import dataclass

from .light_source import LightSource


class Type(Enum):
    HALOGEN = "Halogen"
    INCANDESCENT = "Incandescent"
    OTHER = "Other"


@dataclass
class Filament(LightSource):
    type: Optional[Type] = None



from typing import Optional

from pydantic.dataclasses import dataclass

from .boolean_annotation import BooleanAnnotation
from .comment_annotation import CommentAnnotation
from .double_annotation import DoubleAnnotation
from .file_annotation import FileAnnotation
from .list_annotation import ListAnnotation
from .long_annotation import LongAnnotation
from .map_annotation import MapAnnotation
from .tag_annotation import TagAnnotation
from .term_annotation import TermAnnotation
from .timestamp_annotation import TimestampAnnotation


@dataclass
class StructuredAnnotations:
    boolean_annotation: Optional[BooleanAnnotation] = None
    comment_annotation: Optional[CommentAnnotation] = None
    double_annotation: Optional[DoubleAnnotation] = None
    file_annotation: Optional[FileAnnotation] = None
    list_annotation: Optional[ListAnnotation] = None
    long_annotation: Optional[LongAnnotation] = None
    map_annotation: Optional[MapAnnotation] = None
    tag_annotation: Optional[TagAnnotation] = None
    term_annotation: Optional[TermAnnotation] = None
    timestamp_annotation: Optional[TimestampAnnotation] = None
    xml_annotation: Optional[str] = None



from dataclasses import field
from datetime import datetime
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .simple_types import PlateAcquisitionID, PositiveInt
from .well_sample_ref import WellSampleRef


@dataclass
class PlateAcquisition:
    id: PlateAcquisitionID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None
    end_time: Optional[datetime] = None
    maximum_field_count: Optional[PositiveInt] = None
    name: Optional[str] = None
    start_time: Optional[datetime] = None
    well_sample_ref: List[WellSampleRef] = field(default_factory=list)



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .binary_file import BinaryFile
from .type_annotation import TypeAnnotation

_no_default = object()


@dataclass
class FileAnnotation(TypeAnnotation):
    binary_file: BinaryFile = _no_default  # type: ignore
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    description: Optional[str] = None

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.binary_file is _no_default:
            raise TypeError(
                "__init__ missing 1 required argument: 'binary_file'"
            )



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .annotation_ref import AnnotationRef
from .simple_types import Hex40, NonNegativeInt, UnitsLength, UnitsTime


@dataclass
class Plane:
    the_c: NonNegativeInt
    the_t: NonNegativeInt
    the_z: NonNegativeInt
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    delta_t: Optional[float] = None
    delta_t_unit: Optional[UnitsTime] = UnitsTime("s")
    exposure_time: Optional[float] = None
    exposure_time_unit: Optional[UnitsTime] = UnitsTime("s")
    hash_sha1: Optional[Hex40] = None
    position_x: Optional[float] = None
    position_x_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    position_y: Optional[float] = None
    position_y_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    position_z: Optional[float] = None
    position_z_unit: Optional[UnitsLength] = UnitsLength("reference frame")



from dataclasses import field
from enum import Enum
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .affine_transform import AffineTransform
from .annotation_ref import AnnotationRef
from .simple_types import Color, NonNegativeInt, ShapeID, UnitsLength


class FillRule(Enum):
    EVEN_ODD = "EvenOdd"
    NON_ZERO = "NonZero"


class FontStyle(Enum):
    BOLD = "Bold"
    BOLD_ITALIC = "BoldItalic"
    ITALIC = "Italic"
    NORMAL = "Normal"


class FontFamily(Enum):
    CURSIVE = "cursive"
    FANTASY = "fantasy"
    MONOSPACE = "monospace"
    SANSSERIF = "sans-serif"
    SERIF = "serif"


@dataclass
class Shape:
    id: ShapeID
    annotation_ref: List[AnnotationRef] = field(default_factory=list)
    fill_color: Optional[Color] = None
    fill_rule: Optional[FillRule] = None
    font_family: Optional[FontFamily] = None
    font_size: Optional[NonNegativeInt] = None
    font_size_unit: Optional[UnitsLength] = UnitsLength("pt")
    font_style: Optional[FontStyle] = None
    locked: Optional[bool] = None
    stroke_color: Optional[Color] = None
    stroke_dash_array: Optional[str] = None
    stroke_width: Optional[float] = None
    stroke_width_unit: Optional[UnitsLength] = UnitsLength("pixel")
    text: Optional[str] = None
    the_c: Optional[NonNegativeInt] = None
    the_t: Optional[NonNegativeInt] = None
    the_z: Optional[NonNegativeInt] = None
    transform: Optional[AffineTransform] = None



from enum import Enum
from typing import Optional

from pydantic.dataclasses import dataclass

from .settings import Settings
from .simple_types import ObjectiveID


class Medium(Enum):
    AIR = "Air"
    GLYCEROL = "Glycerol"
    OIL = "Oil"
    OTHER = "Other"
    WATER = "Water"


_no_default = object()


@dataclass
class ObjectiveSettings(Settings):
    id: ObjectiveID = _no_default  # type: ignore
    correction_collar: Optional[float] = None
    medium: Optional[Medium] = None
    refractive_index: Optional[float] = None

    # hack for dataclass inheritance with non-default args
    # https://stackoverflow.com/a/53085935/
    def __post_init__(self):
        if self.id is _no_default:
            raise TypeError("__init__ missing 1 required argument: 'id'")



from typing import Optional

from pydantic.dataclasses import dataclass

from .simple_types import NonNegativeInt, UniversallyUniqueIdentifier


@dataclass
class TiffData:
    uuid: Optional[
        dataclass(
            type(
                "UUID",
                (),
                {
                    "__annotations__": {
                        "file_name": str,
                        "value": UniversallyUniqueIdentifier,
                    }
                },
            )
        )
    ]
    first_c: Optional[NonNegativeInt] = 0
    first_t: Optional[NonNegativeInt] = 0
    first_z: Optional[NonNegativeInt] = 0
    ifd: Optional[NonNegativeInt] = 0
    plane_count: Optional[NonNegativeInt] = None



import importlib
import os
import sys
from glob import glob
from xmlschema.validators.exceptions import XMLSchemaValidationError
import pytest

from ome_autogen import convert_schema
from ome_types import from_xml

this_dir = os.path.dirname(__file__)


@pytest.fixture(scope="session")
def model(tmpdir_factory):
    target_dir = tmpdir_factory.mktemp("test_model")
    xsd = os.path.join(this_dir, "ome-2016-06.xsd")
    convert_schema(url=xsd, target_dir=target_dir)
    sys.path.insert(0, os.path.dirname(str(target_dir)))
    model_name = os.path.basename(str(target_dir))
    model_module = importlib.import_module(model_name)
    return model_module


@pytest.mark.parametrize(
    "xml", glob(os.path.join(this_dir, "*.xml")), ids=lambda x: os.path.basename(x)
)
def test_convert_schema(model, xml):
    if "bad.xml" in xml:
        with pytest.raises(XMLSchemaValidationError):
            assert from_xml(xml, model.OME)
    else:
        assert from_xml(xml, model.OME)
